from canvasapi.course import Course
from canvasapi.file import File
from canvasapi.folder import Folder

from .util import get_canvas_object
from ..resources import FileData, FileInfo
from ..our_logging import get_logger

logger = get_logger()

from ..our_logging import get_logger

logger = get_logger()
DEFAULT_CANVAS_FOLDER = 'deployed_files'


# Keeping for Checksums retrieval
def get_file(course: Course, name: str) -> File:
    return get_canvas_object(course.get_files, 'display_name', name)


def get_canvas_folder(course: Course, folder_name: str, parent_folder_path="") -> Folder:
    """
    Retrieves an object representing a digital folder in Canvas.
    If the folder does not exist, it is created.
    """
    folder = get_canvas_object(course.get_folders, 'name', folder_name)
    if folder is not None:
        return folder

    logger.debug(f"Creating folder: {folder_name}")
    return course.create_folder(name=folder_name, parent_folder_path=parent_folder_path, hidden=True)


def deploy_file(course: Course, data: FileData) -> tuple[FileInfo, None]:
    lock_at = data.get('lock_at')
    unlock_at = data.get('unlock_at')

    canvas_folder = data.get('canvas_folder') or DEFAULT_CANVAS_FOLDER
    folder = get_canvas_folder(course, canvas_folder)
    file_id = folder.upload(data['path'])[1]['id']
    file = course.get_file(file_id)

    # Update the file with lock_at and unlock_at if provided
    if lock_at or unlock_at:
        file.update(lock_at=lock_at, unlock_at=unlock_at)

    file_object_info: FileInfo = {
        'id': file.id,
        'title': file.display_name,
        'uri': f'/files/{file.id}'
    }

    return file_object_info, None



from canvasapi.course import Course

from .util import update_group_name_to_id
from ..resources import AssignmentInfo


def deploy_assignment(course: Course, assignment_info: dict) -> tuple[AssignmentInfo, None]:
    assignment_id = assignment_info["canvas_id"]

    update_group_name_to_id(course, assignment_info)

    # TODO - update group_category (name) to group_category_id
    #  Is this necessary to support?

    if assignment_id:
        canvas_assignment = course.get_assignment(assignment_id)
        canvas_assignment.edit(assignment=assignment_info)
    else:
        canvas_assignment = course.create_assignment(assignment=assignment_info)

    assignment_object_info: AssignmentInfo = {
        'id': canvas_assignment.id,
        'title': canvas_assignment.name,
        'uri': f'/courses/{course.id}/assignments/{canvas_assignment.id}',

        # Following fields have been observed to be missing in some cases
        'url': canvas_assignment.html_url if hasattr(canvas_assignment, 'html_url') else None
    }

    return assignment_object_info, None


def deploy_shell_assignment(course: Course, assignment_info: dict) -> tuple[AssignmentInfo, None]:
    shell_assignment_info = assignment_info.copy()
    shell_assignment_info[
        'description'] = "<p>This is a shell assignment created to break a dependency cycle. The full content will be deployed later.</p>"

    return deploy_assignment(course, shell_assignment_info)



from canvasapi.course import Course


def update_group_name_to_id(course: Course, data: dict):
    if 'assignment_group' in data:
        data['assignment_group_id'] = _get_group_id(course, data['assignment_group'])


def _get_group_id(course: Course, group_name: str) -> int:
    for group in course.get_assignment_groups():
        if group.name == group_name:
            return group.id

    # Group not found - create it
    # TODO - make assignment_group it's own resource
    #  in which case we would throw an error heres
    return course.create_assignment_group(name=group_name).id


def get_canvas_object(course_getter, attr_name, attr):
    objects = course_getter()
    for obj in objects:
        if obj.__getattribute__(attr_name) == attr:
            return obj
    return None


class ResourceNotFoundException(Exception):
    pass



import json
import sys


class DeploymentReport:
    def __init__(self, output_file: str):
        self.output_file = output_file
        self.report = {
            "deployed_content": [],
            "content_to_review": [],
            "error": ""
        }

    def add_deployed_content(self, rtype: str, content_name: str, content_url: str = None):
        self.report["deployed_content"].append((rtype, content_name, content_url))

    def get_deployed_content(self):
        return self.report["deployed_content"]

    def add_content_to_review(self, quiz_name: str, link_to_quiz: str):
        self.report["content_to_review"].append([quiz_name, link_to_quiz])

    def get_content_to_review(self):
        return self.report["content_to_review"]

    def add_error(self, error: Exception):
        error_type = type(error).__name__
        error_msg = str(error)
        self.report["error"] = f"{error_type}: {error_msg}"

    def save_report(self):
        if self.output_file:
            with open(self.output_file, 'w') as f:
                f.write(json.dumps(self.report, indent=4))

    def print_report(self):
        if self.report['deployed_content']:
            groups = {}
            for rtype, rid, url in self.report['deployed_content']:
                if url not in groups:
                    groups[url] = []
                groups[url].append((rtype, rid))

            print(' Deployed Content '.center(60, '-'))
            for url, resources in groups.items():
                resources_str = ', '.join(rid for _, rid in resources)
                print(f'{resources_str}: {url}')

        if self.report['content_to_review']:
            print(' Content to Review '.center(60, '-'))
            for name, url in self.report['content_to_review']:
                print(f'{name}: {url}')

        if self.report['error']:
            print(file=sys.stderr)
            print(self.report['error'], file=sys.stderr)



from canvasapi.course import Course

from ..resources import ModuleInfo, ModuleItemInfo


def deploy_module_item(course: Course, module_item_data: dict) -> tuple[ModuleItemInfo, None]:
    canvas_module = course.get_module(module_item_data['module_id'])
    if canvas_module is None:
        raise ValueError(f'Unable to find module {module_item_data["module_id"]}')

    if module_item_data['canvas_id'] is not None and (
            module_item := canvas_module.get_module_item(module_item_data['canvas_id'])):
        module_item.edit(module_item=module_item_data)
    else:
        module_item = canvas_module.create_module_item(module_item=module_item_data)

    return ModuleItemInfo(
        id=module_item.id,
        module_id=module_item.module_id,
        uri=f'/courses/{course.id}#module_{canvas_module.id}',
        url=f'{course.canvas._Canvas__requester.original_url}/courses/{course.id}#module_{canvas_module.id}'
    ), None


def deploy_module(course: Course, module_data: dict) -> tuple[ModuleInfo, None]:
    module_id = module_data["canvas_id"]

    if module_id:
        canvas_module = course.get_module(module_id)
        if 'published' not in module_data:
            module_data['published'] = canvas_module.published
        canvas_module.edit(module=module_data)
    else:
        canvas_module = course.create_module(module=module_data)

    module_object_info: ModuleInfo = {
        'id': canvas_module.id,
        'title': canvas_module.name,
        'uri': f'/courses/{course.id}#module_{canvas_module.id}',
        'url': f'{course.canvas._Canvas__requester.original_url}/courses/{course.id}'
    }

    return module_object_info, None



from pathlib import Path
from bs4 import Tag
from .processing_context import get_current_file


def get_tag_source_file(tag: Tag) -> Path | None:
    if tag and (source := tag.get('data-source')):
        return Path(source)

    for parent in tag.parents:
        if source := parent.get('data-source'):
            return Path(source)

    return None


def get_file_path(tag: Tag) -> Path:
    if source := get_tag_source_file(tag):
        return source

    if current := get_current_file():
        return current

    raise ValueError("No file context available")


def format_tag(tag: Tag, max_length: int = 80) -> str:
    if not tag:
        return "<unknown tag>"

    tag_name = tag.name
    attrs = tag.attrs or {}

    if not attrs:
        return f"<{tag_name}>"

    # Build attribute strings
    attr_strs = [f'{k}="{" ".join(str(v) for v in val) if isinstance(val, list) else val}"' for k, val in attrs.items()]
    attrs_formatted = ' '.join(attr_strs)
    has_children = bool(tag.contents)

    # Build full tag
    full_tag = f"<{tag_name} {attrs_formatted}>" if has_children else f"<{tag_name} {attrs_formatted} />"

    if len(full_tag) <= max_length:
        return full_tag

    # Truncate if too long (show first few attributes, then indicate more, thanks to Claude)
    shown_attrs = []
    for attr_str in attr_strs[:3]:
        shown_attrs.append(attr_str)
        if len(f"<{tag_name} {' '.join(shown_attrs)} ...>") > max_length - 15:
            shown_attrs.pop()
            break

    remaining = len(attr_strs) - len(shown_attrs)
    if shown_attrs:
        return f"<{tag_name} {' '.join(shown_attrs)} ... +{remaining} more>"
    return f"<{tag_name}> ({len(attr_strs)} attributes)"


def validate_required_attribute(tag: Tag, attr_name: str, tag_display_name: str | None = None) -> str:
    if not (value := tag.get(attr_name)):
        display_name = tag_display_name if tag_display_name else tag.name
        raise ValueError(f'Required field "{attr_name}" missing from {display_name} tag {format_tag(tag)}\n  in {get_file_path(tag)}')
    return value



from .checksums import MD5Sums
from ..our_logging import get_logger

logger = get_logger()


def migrate(course, md5s: MD5Sums):
    """Update the md5 data to match the latest schema"""
    logger.info('Migrating cached data')

    # Module Item -> Module ID map (0.6.6)
    item_id_map = {
        module_item.id: module_item.module_id
        for module in course.get_modules()
        for module_item in module.get_module_items()
    }

    # Override -> Assignment ID map (0.6.6)
    assignment_id_map = {
        override.id: override.assignment_id
        for assignment in course.get_assignments()
        for override in assignment.get_overrides()
    }

    for (rtype, rid), data in md5s._md5s.items():
        # Titles (0.6.2)
        if rtype in ['assignment', 'file', 'module', 'page', 'quiz'] \
                and not data['canvas_info'].get('title'):
            logger.debug(f'Migrating title for {rtype} {rid}')

            canvas_obj = getattr(course, f'get_{rtype}')(data['canvas_info']['id'])
            title = (
                canvas_obj.title if hasattr(canvas_obj, 'title') else
                canvas_obj.name if hasattr(canvas_obj, 'name') else
                canvas_obj.display_name
            )
            md5s._md5s[rtype, rid]['canvas_info']['title'] = title

        elif rtype == 'syllabus' and not data.get('title'):

            logger.debug(f'Migrating title for {rtype} {rid}')
            md5s._md5s[rtype, rid]['canvas_info']['title'] = 'Syllabus'

        # Module Item -> Module ID (0.6.6)
        elif rtype == 'module_item' and not data['canvas_info'].get('module_id'):
            logger.debug(f'Migrating module_id for {rtype} {rid}')

            module_item_id = data['canvas_info'].get('id')
            if module_item_id in item_id_map:
                md5s._md5s[rtype, rid]['canvas_info']['module_id'] = item_id_map[module_item_id]

        # Override -> Assignment ID (0.6.6)
        elif rtype == 'override' and not data['canvas_info'].get('assignment_id'):
            logger.debug(f'Migrating assignment_id for {rtype} {rid}')

            override_id = data['canvas_info'].get('id')
            if override_id in assignment_id_map:
                md5s._md5s[rtype, rid]['canvas_info']['assignment_id'] = assignment_id_map[override_id]



import time
import json
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from collections import defaultdict
from typing import Callable

import pytz
from canvasapi.canvas_object import CanvasObject
from canvasapi.course import Course

from .algorithms import linearize_dependencies
from .announcement import deploy_announcement
from .assignment import deploy_assignment, deploy_shell_assignment
from .checksums import MD5Sums, compute_md5
from .course_settings import deploy_settings
from .file import deploy_file
from .group import deploy_group
from .module import deploy_module, deploy_module_item
from .override import deploy_override
from .page import deploy_page, deploy_shell_page
from .quiz import deploy_quiz, deploy_shell_quiz
from .syllabus import deploy_syllabus
from .zip import deploy_zip, predeploy_zip
from ..deployment_report import DeploymentReport
from ..our_logging import get_logger
from ..resources import CanvasResource, iter_keys

from .migration import migrate

logger = get_logger()

PREDEPLOYERS: dict[str, Callable[[dict, Path], dict]] = {
    'zip': predeploy_zip
}

SHELL_DEPLOYERS = {
    # Current known resources that need shell deployments
    'assignment': deploy_shell_assignment,
    'page': deploy_shell_page,
    'quiz': deploy_shell_quiz
}

DEPLOYERS = {
    'announcement': deploy_announcement,
    'assignment': deploy_assignment,
    'assignment_group': deploy_group,
    'course_settings': deploy_settings,
    'file': deploy_file,
    'module': deploy_module,
    'module_item': deploy_module_item,
    'override': deploy_override,
    'page': deploy_page,
    'quiz': deploy_quiz,
    'syllabus': deploy_syllabus,
    'zip': deploy_zip
}


def get_dependencies(resources: dict[tuple[str, str], CanvasResource]) -> dict[tuple[str, str], list[tuple[str, str]]]:
    """Returns the dependency graph in resources. Adds missing resources to the input dictionary."""
    deps = {}
    missing_resources = []
    for key, resource in resources.items():
        deps[key] = []
        text = json.dumps(resource)
        for _, rtype, rid, _ in iter_keys(text):
            resource_key = (rtype, rid)
            deps[key].append(resource_key)
            if resource_key not in resources:
                missing_resources.append(resource_key)

    for rtype, rid in missing_resources:
        resources[rtype, rid] = CanvasResource(type=rtype, id=rid, data={}, content_path='')

    return deps


def make_iso(date: datetime | str | None, time_zone: str) -> str:
    if isinstance(date, datetime):
        return datetime.isoformat(date)

    if isinstance(date, str):
        try_formats = [
            "%b %d, %Y, %I:%M %p",
            "%b %d %Y %I:%M %p",
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%M:%S%z"
        ]

        for format_str in try_formats:
            try:
                parsed_date = datetime.strptime(date, format_str)
                if parsed_date.tzinfo:
                    return datetime.isoformat(parsed_date)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Invalid date format: {date}")

        to_zone = pytz.timezone(time_zone)
        localized_date = to_zone.localize(parsed_date)
        return datetime.isoformat(localized_date)

    raise TypeError("Date must be a datetime object or a string")


def fix_dates(data: dict, time_zone: str, resource: CanvasResource):
    for attr in ['due_at', 'unlock_at', 'lock_at', 'show_correct_answers_at']:
        if (val := data.get(attr)) is None:
            continue

        try:
            dt = datetime.fromisoformat(make_iso(val, time_zone))
            data[attr] = dt.astimezone(pytz.utc).isoformat()
        except (ValueError, TypeError) as e:
            raise ValueError(
                f"Invalid date format for {resource['type']} {resource['id']}\n  {attr}: '{val}'\n  in {resource['content_path']}") from e


def predeploy_resources(resources: dict, timezone: str, tmpdir: Path):
    for resource in resources.values():
        if (data := resource.get('data')) is None:
            continue

        fix_dates(data, timezone, resource)

        rtype = resource['type']
        if predeploy := PREDEPLOYERS.get(rtype):
            logger.debug(f'Predeploying {rtype} {data}')
            resource['data'] = predeploy(data, tmpdir)


def identify_modified_or_outdated(
        resources: dict[tuple[str, str], CanvasResource],
        linearized_resources: list[tuple[tuple[str, str], bool]],
        resource_dependencies: dict[tuple[str, str], list[tuple[str, str]]],
        md5s: MD5Sums
) -> dict[tuple[str, str], tuple[str, CanvasResource]]:
    """
    A resource is modified or outdated if:
        - It is new
        - It has changed its own data
        - It depends on another resource with a new ID (a file)

    Returns:
        dict: A dictionary mapping resource keys to their current MD5 and resource data.
            - Key: (resource_key, is_shell)
                - resource_key: (type, id)
                    - type: str, the resource type (e.g., 'assignment', 'page', etc.)
                    - id: str, the resource identifier that the user assigned (not the Canvas ID)
                - is_shell: bool, indicating if this is a shell deployment
                    - Unfortunately needed to handle shell deployments properly, otherwise shell deployments are
                      overwritten by full deployments of the same resource.
            - Value: (current_md5, resource)
                - current_md5: str, the current MD5 checksum of the resource data
                - resource: CanvasResource, the resource data itself
    """
    modified = {}

    for resource_key, is_shell in linearized_resources:
        resource = resources[resource_key]
        if (resource_data := resource.get('data')) is None:
            # Just a resource reference
            continue

        item = (resource['type'], resource['id'])

        stored_md5 = md5s.get_checksum(item)
        current_md5 = compute_md5(resource_data)

        # Attach the Canvas object id (stored as `canvas_id`) to the resource data
        # so deployment can detect whether to create a new item or update an existing one.
        resource['data']['canvas_id'] = md5s.get_canvas_info(item).get('id') if md5s.has_canvas_info(item) else None

        if stored_md5 is None:
            # New resource that needs to be deployed
            modified[resource_key, is_shell] = current_md5, resource
            continue

        if is_shell:
            # Shell deployments only needed for new resources
            # stored_md5 is not None, so the resource is not new
            # so we can skip
            continue

        if stored_md5 != current_md5:
            # Changed data, need to deploy
            logger.debug(f'MD5 {resource_key}: {current_md5} vs {stored_md5}')
            modified[resource_key, is_shell] = current_md5, resource
            continue

        for dep_type, dep_name in resource_dependencies[resource_key]:
            if dep_type in ['file', 'zip'] and (dep_type, dep_name) in modified:
                modified[resource_key, is_shell] = current_md5, resource
                break

    return modified


def update_links(md5s: MD5Sums, data: dict, resource_objs: dict, current_resource: CanvasResource) -> dict:
    text = json.dumps(data)

    for key, rtype, rid, field in iter_keys(text):
        canvas_info = resource_objs.get((rtype, rid)) or md5s.get_canvas_info((rtype, rid))

        if not canvas_info:
            logger.debug(data)
            raise ValueError(
                f"No canvas info for {rtype} {rid}\n  Referenced in {current_resource['type']} {current_resource['id']}\n  in {current_resource['content_path']}")

        if not (repl_text := canvas_info.get(field)):
            raise ValueError(
                f"Missing field '{field}' in {rtype} {rid}\n  Referenced in {current_resource['type']} {current_resource['id']}\n  in {current_resource['content_path']}")

        text = text.replace(key, str(repl_text))

    return json.loads(text)


def deploy_resource(deployers: dict, course: Course, rtype: str, data: dict, resource: CanvasResource):
    if not (deploy := deployers.get(rtype)):
        raise Exception(f"Unsupported resource type {rtype} {resource['id']}\n  in {resource['content_path']}")

    try:
        deployed, info = deploy(course, data)
    except Exception as e:
        raise Exception(
            f"Error deploying {rtype} {resource['id']}\n  {type(e).__name__}: {e}\n  in {resource['content_path']}") from e

    if not deployed:
        raise Exception(f"Deployment returned None for {rtype} {resource['id']}\n  in {resource['content_path']}")

    return deployed, info


def deploy_to_canvas(course: Course, timezone: str, resources: dict[tuple[str, str], CanvasResource],
                     report: DeploymentReport, dryrun=False):
    resource_dependencies = get_dependencies(resources)
    logger.debug(f'Dependency graph: {resource_dependencies}')

    resource_order = linearize_dependencies(resource_dependencies, list(SHELL_DEPLOYERS.keys()))
    logger.debug(f'Linearized dependencies: {resource_order}')

    logger.info('Preparing resources for deployment to Canvas')

    with MD5Sums(course) as md5s, TemporaryDirectory() as tmpdir:
        migrate(course, md5s)

        tmpdir = Path(tmpdir)

        predeploy_resources(resources, timezone, tmpdir)

        to_deploy = identify_modified_or_outdated(resources, resource_order, resource_dependencies, md5s)
        total = len(to_deploy)
        if not total:
            logger.info('No resources to deploy')
            return

        # Summary by type
        grouped = defaultdict(int)
        for (rtype, _), _ in to_deploy.keys():
            grouped[rtype] += 1

        logger.info('=' * 80)

        logger.info(f'Resources to deploy: {total}')
        max_len = max(len(rtype) for rtype in grouped)
        for rtype, count in sorted(grouped.items()):
            logger.info(f'  {rtype:{max_len}}  {count:>3}')

        logger.info('=' * 80)

        if dryrun:
            logger.info('Dry run - no resources deployed')
            return

        logger.info('Deploying resources to Canvas')
        start_time = time.perf_counter()
        resource_objs: dict[tuple[str, str], CanvasObject] = {}
        index_width = len(str(total))
        for index, ((resource_key, is_shell), (current_md5, resource)) in enumerate(to_deploy.items(), start=1):
            rtype, rid = resource_key

            if (resource_data := resource.get('data')) is not None:
                shell_tag = '(shell) ' if is_shell else ''
                logger.info(f'[{index:>{index_width}}/{total}] {shell_tag}{rtype:{max_len}}  {rid}')

                if is_shell:
                    canvas_obj_info, info = deploy_resource(SHELL_DEPLOYERS, course, rtype, resource_data, resource)
                    resource['data']['canvas_id'] = canvas_obj_info.get('id') if canvas_obj_info else None
                else:
                    resource_data = update_links(md5s, resource_data, resource_objs, resource)
                    canvas_obj_info, info = deploy_resource(DEPLOYERS, course, rtype, resource_data, resource)

                if canvas_obj_info:
                    resource_objs[resource_key] = canvas_obj_info
                    if url := canvas_obj_info.get('url'):
                        report.add_deployed_content(rtype, rid, url)

                if info:
                    report.add_content_to_review(*info)

                md5s[resource_key] = {"checksum": current_md5, "canvas_info": canvas_obj_info}

        elapsed = time.perf_counter() - start_time
        logger.info(f'Deployment complete - {total} resources in {elapsed:.1f}s')



from canvasapi.course import Course

from ..resources import CourseSettings, CourseSettingsInfo


class CourseObj:
    def __init__(self, course_id: int):
        self.course_id = int
        self.uri = f'/courses/{course_id}'


def deploy_settings(course: Course, data: CourseSettings) -> tuple[CourseSettingsInfo, None]:
    course.update(course={
        'name': data['name'],
        'course_code': data['code'],
        'image_id': int(data['image']) if data.get('image') else None
        # TODO: syllabus field
    })

    settings_object_info: CourseSettingsInfo = {
        'id': str(course.id)
    }

    return settings_object_info, None



from pathlib import Path
from contextvars import ContextVar

_file_stack: ContextVar[list[Path] | None] = ContextVar('file_stack', default=None)


class FileContext:
    """Tracks current file being processed for error reporting."""

    def __init__(self, file_path: Path):
        self.file_path = file_path.resolve()
        self.reset_token = None

    def __enter__(self):
        current = _file_stack.get(None) or []
        self.reset_token = _file_stack.set(current + [self.file_path])
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _file_stack.reset(self.reset_token)


def get_current_file() -> Path | None:
    """Returns the file currently being processed."""
    files = _file_stack.get(None)
    return files[-1] if files else None


def get_file_stack() -> list[Path]:
    """Returns the complete file processing chain."""
    return _file_stack.get(None) or []



from canvasapi.assignment import Assignment
from canvasapi.course import Course

from ..resources import OverrideInfo


def _get_assignment(course: Course, override_info: dict) -> Assignment:
    rtype = override_info.get('rtype')

    # Not necessarily needed, as it never occurs in practice, but just to be safe.
    # Only assignments and quizzes parse overrides.
    if rtype not in ['assignment', 'quiz']:
        raise ValueError(f"Invalid override rtype: {rtype}. Must be 'assignment' or 'quiz'.")

    assignment_id = int(override_info.get('assignment_id'))

    if rtype == 'quiz':
        # Quizzes are unique in that they have two IDs: the quiz ID and the assignment ID.
        # To get and modify a quizzes override, we need the assignment ID which we do not track directly.
        # This step gets the quiz object to retrieve the assignment ID.

        quiz = course.get_quiz(assignment_id)
        assignment_id = quiz.assignment_id

    return course.get_assignment(assignment_id)


def deploy_override(course: Course, override_info: dict) -> tuple[OverrideInfo, None]:
    assignment: Assignment = _get_assignment(course, override_info)

    if cid := override_info.get('canvas_id'):
        override = assignment.get_override(cid)
        override.edit(assignment_override=override_info)
    else:
        override = assignment.create_override(assignment_override=override_info)

    override_object_info: OverrideInfo = {
        'id': override.id,
        'assignment_id': assignment.id,
    }

    return override_object_info, None
